/****************************************************************************
**
** Copyright (C) 2017 The Qt Company Ltd.
** Contact: http://www.qt-project.org/legal
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:BSD$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** BSD License Usage
** Alternatively, you may use this file under the terms of the BSD license
** as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of The Qt Company Ltd nor the names of its
**     contributors may be used to endorse or promote products derived
**     from this software without specific prior written permission.
**
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include <QtCore/QUrl>
#include <QtCore/QDebug>
#include <QSettings>

#include <QtGui/QGuiApplication>

#include <QtQml/QQmlApplicationEngine>
#include <QQmlContext>
#include <QQmlComponent>
#include <QFile>
#include <QProcess>
#include <QThread>
#include <QTimer>
#include <QDir>
#include <unistd.h>
#include <signal.h>
#include "main.h"
#include "nm_dbus_server.h"

Gpqml* gpqml=NULL;
QQmlApplicationEngine *appEngine=NULL;
QQuickWindow *wq=NULL;
QString root_path;
monitor_thread *th;
nm_dbus_server *iface=NULL;


monitor_thread::monitor_thread():QThread()
{
	zygoterunning = false;
	QObject::connect(gpqml,&Gpqml::sig_zygote_running,this , &monitor_thread::on_sig_zygote_running);
}	
	
void monitor_thread::run()
{
	QString cmd="/bin/pidof android.hardware.graphics.composer@2.1-service";
	QString cmd4="/bin/pidof android.hardware.graphics.composer@2.4-service";
	QString post_file = "/usr/bin/run-compositor-after.sh";
	QString s;
	QProcess p;
	int pid_i = 0;
	bool iscomposer4exists=false;

	{
		if (QFile(post_file).exists()){
			p.start("/bin/sh",QStringList() << "-c" << post_file);
			qDebug() << "running " << post_file;
			if (p.waitForFinished()){
			}
		}
	}
	if (QFile("/vendor/bin/hw/android.hardware.graphics.composer@2.4-service").exists())
		iscomposer4exists = true;
	qDebug() << "iscomposer4exists" << iscomposer4exists;
	while(1){		
		p.start("/bin/sh",QStringList() << "-c" << (iscomposer4exists?cmd4:cmd));
		if (p.waitForFinished()){
		}
		s = p.readAll();
		//qDebug() << "pid is " << s;
		if (s!=""){
			int pid;
			bool zygoterunningi;
			
			if (!pid_i)
				pid_i = s.toInt();
			pid  = s.toInt();
			if (pid != pid_i)
				::exit (-1);

			
			m.lock();
			zygoterunningi = zygoterunning;
			m.unlock();
			
			if (0/*!zygoterunningi*/){//disable for the moment
				bool isrunning = Gpqml::if_service_active("zygote");

				//qDebug() << "zygoterunning " << isrunning;
				if (isrunning != zygoterunningi){
					Gpqml::set_property_bool("zygoterunning",isrunning);
					if (!isrunning)
						iface->androidgui_status(0);//running
				}
				m.lock();
				zygoterunning = isrunning;
				m.unlock();
			}
			usleep(2000000);
			continue;
		}
		usleep(2000000);
		//::exit (-1);
	}
}
int monitor_thread::get_zygote_status()
{
	bool zygoterunningi;

	m.lock();
	zygoterunningi = zygoterunning;
	m.unlock();

	qDebug() << "zygoterunning " << zygoterunningi;
	return (zygoterunningi)?2:0;
}
void monitor_thread::start_zygote(bool isrun)
{
	bool zygoterunningi = get_zygote_status();

	if (isrun && !zygoterunningi){
		gpqml->run_process("systemctl start zygote","zygote","",-1);
	}
	else if(!isrun && zygoterunningi){
		gpqml->run_process("systemctl start zygote","","",-1);
	}
}
int  monitor_thread::raise_window(const QString & winname,const QString & appname,int display_id)
{
	QVariant returnedValue;
	//gpqml->sig_raise_window(winname,appname,display_id);
	QMetaObject::invokeMethod(wq, "raise_window",Q_RETURN_ARG(QVariant, returnedValue),
							  Q_ARG(QVariant, winname), Q_ARG(QVariant, appname), Q_ARG(QVariant, display_id));
	return returnedValue.toInt();
}
void monitor_thread::compositor_wakeup()
{
	gpqml->sig_compositor_wakeup();
}
void monitor_thread::take_screnshot()
{
	gpqml->sig_take_screenshot();
}
void monitor_thread::on_sig_zygote_running(int val)
{
	qDebug() << "val " << val;
	iface->androidgui_status(val);
}
//dbus-send --system --print-reply  --dest=org.editec.compositor /compositor  run.compositor.delete_window string:"premierbetzone.com"
void monitor_thread::delete_window(const QString & winname)
{
	QMetaObject::invokeMethod(wq, "delete_window",Q_ARG(QVariant, winname));
}
//dbus-send --system --print-reply  --dest=org.editec.compositor /compositor  run.compositor.show_panel array:string:"0","-1","-1"
void monitor_thread::show_panel(bool isshow,int display_id,int height)
{
	QMetaObject::invokeMethod(wq, "show_panel",Q_ARG(QVariant, isshow),Q_ARG(QVariant, display_id),Q_ARG(QVariant, height));
}
//dbus-send --system --print-reply=literal  --dest=org.editec.compositor /compositor  run.compositor.update_window_pid_info  int32:-1 string:"premierbetzone" string:""
void monitor_thread::update_window_pid_info(int pid,const QString&url)
{
	gpqml->sig_update_window_pid_info(pid,url);
}
Gpqml::Gpqml():QObject()
{
	gpqml = this;
	fw.addPath(root_path + "/launchers/");
	connect(&fw, SIGNAL(directoryChanged(const QString&)), this,SLOT(on_directoryChanged(const QString&)));
	qDebug() << "Gpqml";
}
void Gpqml::on_directoryChanged(const QString &path)
{
	qDebug() << "changed " << path;
	sig_desktopfiles_changed();
}
void Gpqml::set_brightness(int display_id,int value)
{
    QFile f;
    QString s;

    s = QString::number(value);
    if (display_id == 0){
#if defined(__aarch64__)		
		f.setFileName("/sys/class/leds/lcd-backlight/brightness");
#else
		f.setFileName("/sys/class/backlight/sprd_backlight/brightness");
#endif		
	}
    else
		f.setFileName("/sys/class/leds/lcd-backlight2/brightness");
    f.open(QIODevice::WriteOnly);
    f.write(s.toUtf8().data(),s.length());
    f.close();
    
}
QString Gpqml::get_env(const QString &s)
{
    return qgetenv(s.toUtf8());
}
QString Gpqml::get_config(const QString &s,const QString &def)
{
    QSettings  settings("/etc/pure-qml.conf",QSettings::NativeFormat);

    return settings.value(s,def).toString();
}
void Gpqml::set_prop(const QString &key,const QString &val)
{
	QProcess builder;

	builder.start("/usr/bin/setprop", QStringList() << key << val);
	builder.waitForFinished();
}
bool Gpqml::if_process_exist(const QString &pname)
{
     QProcess builder;
     QString s;
     int     pid=0;

     builder.start("/bin/pidof", QStringList() << pname);
     
     if (!builder.waitForFinished())
      return false;
     s = builder.readAll();
     sscanf(s.toUtf8().data(),"%d",&pid);
	 qDebug() << "pid " << pid;
     if (pid > 0)
      return true;
     
     return false;
}
bool Gpqml::if_service_active(const QString &pname)
{
     QProcess builder;
     QString s;

	 s = QString("/bin/systemctl is-active %1").arg(pname);
	 
     builder.start("/bin/sh", QStringList() << "-c" << s);
	 
     if (!builder.waitForFinished()){
		 qDebug() << "error running status " << pname << builder.errorString();
		 return false;
	 }	 
     s = builder.readAll().trimmed();
     if (s == "active")
      return true;

    return false;
}
int Gpqml::run_android_process(const QString &cmd,const QString &proc_name,const QString &raise_app,int display_id)
{
    QProcess *pp;
	QString cmdi = cmd;
    int nb_try = 10;
    int i;
	int display_id_i=-1;

	//qDebug() << cmd << proc_name << raise_app;
    if (if_process_exist(proc_name))
		return 0;
	if (raise_app != "" && th->raise_window(raise_app,"",-1) >=0)
		return 0;
	if (display_id<=1)
		display_id_i = 0;
	else if (display_id==2){
		display_id_i = get_second_displayid();
	}
	else if (display_id==3){
		display_id_i = get_ext_displayid();
		if (display_id_i < 0)
			display_id_i = 1;
	}
    for (i=0;i < nb_try;i++){//max 5 sec
		pp = new QProcess;
		pp->setProcessChannelMode(QProcess::ForwardedChannels);
		if (display_id_i > 0){
			cmdi = cmd + " --display " + QString::number(display_id_i);
		}
		pp->start(cmdi);
		qDebug() << cmdi;
		qDebug() << pp->readAll();
		if (pp->waitForStarted()){
		}
		usleep(500000);
		if(if_process_exist(proc_name))
			break;//
    }
    return (i != nb_try)?0:-1;
}
int Gpqml::run_process(const QString &cmd,const QString &service_name,const QString &raise_app,int display_id)
{
    QProcess *pp;

	//qDebug() << cmd << service_name << raise_app;
    if (service_name != "" && if_service_active(service_name))
		return 0;
	if (raise_app != "" && th->raise_window(raise_app,"",-1) >=0)
		return 0;

    {
		pp = new QProcess;
		if (display_id >0){
			QProcessEnvironment env = QProcessEnvironment::systemEnvironment();

			env.insert("DISPLAY_ID", QString::number(display_id));
			pp->setProcessEnvironment(env);
		}
		pp->setProcessChannelMode(QProcess::ForwardedChannels);
		pp->start(cmd);		
		qDebug() << pp->readAll();
		if (pp->waitForStarted()){
		}
    }
    return 0;
}
int Gpqml::run_android_process_by_pkg(const QString &pkg)
{
	QProcess *pp;
	QProcess p;
	QString cmdb="am stack list|grep %1|cut -d ' ' -f4";
	QString cmd="am start %1";
	QString s;

	p.start("/bin/sh",QStringList() << "-c" << cmdb.arg(pkg));
    if (p.waitForFinished()){
    }
    s = p.readAll();
    qDebug() << s;
    {
		pp = new QProcess;
		pp->setProcessChannelMode(QProcess::ForwardedChannels);
		pp->start(cmd.arg(s));
		qDebug() << pp->readAll();
		if (pp->waitForStarted()){
		}	
    }
    return 0;
}
int  Gpqml::get_ext_displayid()
{
    int display_id = -1;
    QString cmd="dumpsys display |grep -i displaylink|/bin/sed 's/,/\\n/g'|grep layerStack|head -n1| cut -d ' ' -f3";
    QString s;//uff it is ugly to find ext display_id this way
    QProcess p;

    p.start("/bin/sh",QStringList() << "-c" << cmd);
    if (p.waitForFinished()){
    }
    s = p.readAll();
    qDebug() << s;
    if (s!=""){
	display_id = s.toInt();
    }
    return display_id;
}
int  Gpqml::get_second_displayid()
{
	char s[256];
	int  display_id = 0;
	
	if (getenv("USE_DISP")){
		strncpy(s,getenv("USE_DISP"),sizeof(s)-1);
		if (atol(s) == 3)
			display_id = 1;
	}
	return display_id;	
}
QObject* Gpqml::create_qml_component(const QString& name)
{
    QObject *oq=NULL;
    QString  fpath;

    fpath = root_path  + name + ".qml";
    if (!QFile(fpath).exists()){
        qDebug() << fpath << "does not exists";
        return NULL;
    }
    QQmlComponent *component;
    QQmlContext *context = new QQmlContext(appEngine->rootContext());
    component = new QQmlComponent(appEngine,fpath);
    oq = component->create(context);
    if (!oq){
        for (int i=0; i < component->errors().count();i++){
            qDebug() << "line: " << component->errors().at(i).line() << component->errors().at(i).description();
        }
        qDebug() << "Error: Your qml item can not be created " <<fpath;
        return NULL;
    }

    return oq;
}
QQuickWindow* Gpqml::load_qml_window(const QString &name)
{
    QQuickWindow *wq = NULL;

    //lastwndname = name;
    wq = (QQuickWindow *)create_qml_component(name);
    if (!wq)
	return NULL;
    wq->show();
    qDebug() << "created window " + name << wq;
    return wq;
}
void Gpqml::set_property_bool(const QString &key,bool v)
{
	if (appEngine->rootObjects().size() > 0){
		appEngine->rootObjects()[0]->setProperty(key.toUtf8().data(), v);
    }
}
void Gpqml::sleep(int ms)
{
	usleep(ms*1000);
}
void Gpqml::screenshot(QVariantList wl)
{
	for (int i=0;i < wl.length();i++){
		QQuickWindow *w = (QQuickWindow *)wl[i].value<QQuickWindow *>();
		QImage im = w->grabWindow();
		QString rp = "/home/root/screenshots";
		QDir d(rp);		
		int  id = 0;

		d.mkdir(rp);
		d.setNameFilters(QStringList() << QString("*_D%1.jpg").arg(i+1));
		QFileInfoList list = d.entryInfoList();		
		for (int j = 0; j < list.size(); j++) {
			QStringList l = list[j].baseName().split("_");
			if (l.size() > 0){
				if (l[0].toInt() > id){
					id = l[0].toInt();
				}
			}
		}
		im.save(QString("%1/%2_D%3.jpg").arg(rp).arg(id+1,5,10,QChar('0')).arg(i+1));
	}
}
bool  Gpqml::parse_line(char *buf,char str[10][4096],int min_nb)
{
     int result;
	 char s1[128];
	 char s2[128];
	 char s3[128];

     if(!buf[0] || buf[0]=='#')
		 return false;
     memset(str,0,10*4096);
	 if (min_nb < 4){
		 result = sscanf(buf,"%[_a-zA-Z0-9/]%[\t\" ]%[_,a-zA-Z0-9 /-#\n]%[\t\" ]%[^\t\n\r]",str[0],s1,str[1],s2,str[2]);
		 result-=2;
		 //fprintf(stderr,"buf: <%s> result %d \n",buf,result);
	 }
	 else {
		 result = sscanf(buf,"%[_a-zA-Z0-9/]%[\t\" ]%[_,a-zA-Z0-9 /-#]%[\t\" ]%[_,a-zA-Z0-9#= /-]%[\t\" ]%[^\t\n\r]",str[0],s1,str[1],s2,str[2],s3,str[3]);
		 result-=3;
	 }	 
     return (result>=min_nb)?true:false;
}
QString Gpqml::get_project_name()
{
	char buf[4096];
	char str[10][4096];
	QString name;	
	FILE *fp;
    
	fp = fopen("/home/root/deploy_arm/features.txt","r");
	if (!fp)
		return name;
	while (!feof(fp)){
		char *p = fgets(buf, sizeof(buf),fp);
		
		if (!p)
			return name;	  
		if(!parse_line(buf,str,3))
			continue;
		if (QString::fromLatin1(str[1]) == "project_name"){
			fclose(fp);
			return QString(str[2]);
		}
	}
	fclose(fp);
	return name;
}
//QFileInfoList  list = dr.entryInfoList(QStringList() << "*u-blox*");
QVariantList Gpqml::load_desktop_files(int display_id_show_i)
{
	QVariantList l;
	QDir      dr(root_path + "/launchers/","*.desktop",QDir::Name | QDir::IgnoreCase );
	QFileInfoList  list = dr.entryInfoList(QStringList() << "*.desktop");
	QString project_name = get_project_name();
	
	for (int i=0 ;i<list.count();i++) {
		QSettings s(list.at(i).canonicalFilePath(),QSettings::NativeFormat);
		desktopfile *d;
		int display_id_show;
		QString sl;

		display_id_show = s.value(QString("Desktop Entry/DisplayShow_%1").arg(project_name),-2).toInt();
		if (display_id_show == -2)
			display_id_show = s.value("Desktop Entry/DisplayShow",-1).toInt();
		if (display_id_show >=0 && display_id_show != display_id_show_i)
			continue;
		d = new desktopfile();
		d->name = s.value("Desktop Entry/Name").toString();
		d->icon = s.value("Desktop Entry/Icon").toString();
		d->display_id = s.value("Desktop Entry/Display",-1).toInt();
		d->display_id_show = display_id_show;
		d->appname = s.value("Desktop Entry/Appname","").toString();
		d->pos_x = s.value("Desktop Entry/Pos_x",-1).toInt();
		d->raise_app = s.value(QString("Desktop Entry/RaiseApp_%1").arg(project_name),"").toString();
		if (d->raise_app == "")
			d->raise_app = s.value("Desktop Entry/RaiseApp","").toString();
		d->issuspendbackground = s.value("Desktop Entry/SuspendBackground",false).toBool();
		d->exec_kill = s.value("Desktop Entry/Kill","").toString();
		d->exec_kill_window = s.value("Desktop Entry/KillWindow",false).toBool();
		d->isignoretoppanel = s.value("Desktop Entry/IgnoreTopPanel",false).toBool();
		for (int j=0;j < 10;j++){
			sl = s.value(QString("Desktop Entry/Exec_%1_%2").arg(j+1).arg(project_name),"").toString();
			if (sl == "")
				sl = s.value(QString("Desktop Entry/Exec_%1").arg(j+1),"").toString();
			if (sl != ""){
				d->exec.push_back(sl);
			}
			sl = s.value(QString("Desktop Entry/Exec_a_%1_%2").arg(j+1).arg(project_name),"").toString();
			if (sl == "")
				sl = s.value(QString("Desktop Entry/Exec_a_%1").arg(j+1),"").toString();
			if (sl != ""){
				d->exec_a.push_back(sl);
			}
			sl = s.value(QString("Desktop Entry/Exec_js_%1_%2").arg(j+1).arg(project_name),"").toString();
			if (sl == "")
				sl = s.value(QString("Desktop Entry/Exec_js_%1").arg(j+1),"").toString();
			if (sl != ""){
				d->exec_js.push_back(sl);
			}
		}
		sl = s.value(QString("Desktop Entry/Exec_%1").arg(project_name),"").toString();
		if (sl == "")
			sl = s.value(QString("Desktop Entry/Exec"),"").toString();
		if (sl != ""){
			d->exec.push_back(sl);
		}
		sl = s.value(QString("Desktop Entry/Exec_a_%1").arg(project_name),"").toString();
		if (sl == "")
			sl = s.value(QString("Desktop Entry/Exec_a"),"").toString();
		if (sl != ""){
			d->exec_a.push_back(sl);
		}
		sl = s.value(QString("Desktop Entry/Exec_js_%1").arg(project_name),"").toString();
		if (sl == "")
			sl = s.value(QString("Desktop Entry/Exec_js"),"").toString();
		if (sl != ""){
			d->exec_js.push_back(sl);
		}
		l.push_back(QVariant::fromValue(d));
	}
	
	return l;
}
void Gpqml::suspend_process(int pid,bool issuspend)
{
	kill(pid,issuspend?SIGSTOP:SIGCONT);
}
void Gpqml::kill_process(int pid)
{
	kill(pid,SIGKILL);
}
int main(int argc, char *argv[])
{
    // ShareOpenGLContexts is needed for using the threaded renderer
    // on Nvidia EGLStreams
    QString qmlpath;
    QCoreApplication::setAttribute(Qt::AA_ShareOpenGLContexts, true);
    QGuiApplication app(argc, argv);
    QString path = QCoreApplication::applicationDirPath()+"/";
    QByteArray use_disp;
	QString tpm_type;
	

    appEngine = new QQmlApplicationEngine();
    //qmlpath = path;
    app.setOrganizationName("editec");
    app.setApplicationName("compositor");
    use_disp = qgetenv("USE_DISP");
#if defined(__aarch64__)
    root_path = qmlpath = "/usr/lib64/qml/pure-qml/";
	tpm_type = "ts10";
#else
	root_path = qmlpath = "/usr/lib/qml/pure-qml/";
	tpm_type = "llt5";
#endif
    appEngine->addImportPath(qmlpath);
    qmlRegisterType<Gpqml>("editec.base", 1, 0, "Gpqml");
	qmlRegisterType<desktopfile>("editec.base", 1, 0, "DesktopFile");
    //appEngine.setInitialProperties(QVariantMap{{"use_disp", use_disp}});
    appEngine->load(qmlpath + "main.qml");
    if (appEngine->rootObjects().size() > 0){
		appEngine->rootObjects()[0]->setProperty("use_disp", use_disp);
		appEngine->rootObjects()[0]->setProperty("tpm_type", tpm_type);
		wq = (QQuickWindow *)appEngine->rootObjects()[0];
		//appEngine->rootContext()->setContextProperty("gpqml", gpqml);
    }
	
	th = new monitor_thread();
	QTimer::singleShot(2000,NULL, [th]() {th->start();} );

	iface = nm_dbus_server_init(th);
	if (!iface)
		return -1;
    return app.exec();
}
